/*
 *  Copyright 2005-2006 The Apache Software Foundation or its licensors, as applicable.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
/** 
 * @author Andrey Chernyshev
 * @version $Revision: 1.1.2.1.4.4 $
 */  

#ifndef _ATOMICS_H_
#define _ATOMICS_H_

#include "jni.h"
#include "Class.h"
#include "vm_synch.h"

JNIEXPORT jboolean compareAndSetObjectField
  (JNIEnv * env, jobject self, jobject obj, jobject fieldID, jobject expected, jobject value);
      
JNIEXPORT jboolean compareAndSetBooleanField
  (JNIEnv * env, jobject self, jobject obj, jobject fieldID, jboolean expected, jboolean value);
                  
JNIEXPORT jboolean compareAndSetIntField
  (JNIEnv * env, jobject self, jobject obj, jobject fieldID, jint expected, jint value);
                  
JNIEXPORT jboolean compareAndSetLongField
  (JNIEnv * env, jobject self, jobject obj, jobject fieldID, jlong expected, jlong value);
               
JNIEXPORT jboolean compareAndSetObjectArray
(JNIEnv * env, jobject self, jobjectArray array, jint index, jobject expected, jobject value);

JNIEXPORT jboolean compareAndSetBooleanArray
(JNIEnv * env, jobject self, jbooleanArray array, jint index, jboolean expected, jboolean value);

JNIEXPORT jboolean compareAndSetIntArray
(JNIEnv * env, jobject self, jintArray array, jint index, jint expected, jint value);

JNIEXPORT jboolean compareAndSetLongArray
(JNIEnv * env, jobject self, jlongArray array, jint index, jlong expected, jlong value);

//void MemoryReadWriteBarrier();

#if defined (PLATFORM_POSIX) 
    #if defined (_IPF_)
        inline void MemoryReadWriteBarrier() {
            asm volatile ("mf" ::: "memory");
        }
        inline void MemoryWriteBarrier() {
            asm volatile ("mf" ::: "memory");
        }
    #else 
        inline void MemoryReadWriteBarrier() {
            __asm__("mfence");
        } 
        inline void MemoryWriteBarrier() {
            __asm__("sfence");
        } 
    #endif
#else
    #if defined (_IPF_)
        inline void MemoryReadWriteBarrier() {
            __asm mf;
        }
        inline void MemoryWriteBarrier() {
            __asm mf;
        }
    #else
        inline void MemoryReadWriteBarrier() {
            __asm mfence;
        }
        inline void MemoryWriteBarrier() {
            __asm sfence;
        }
    #endif
#endif

#endif // _ATOMICS_H_

